import { TinyColor } from '@ctrl/tinycolor';
import fs from 'fs';
import fileType from 'file-type';
import Jimp from 'jimp';
import gifWrap from 'gifwrap';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __awaiter(thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

var TimeboxEvoMessageArray = (function (_super) {
    __extends(TimeboxEvoMessageArray, _super);
    function TimeboxEvoMessageArray(items) {
        return _super.apply(this, items) || this;
    }
    TimeboxEvoMessageArray.create = function () {
        return Object.create(TimeboxEvoMessageArray.prototype);
    };
    TimeboxEvoMessageArray.prototype.asBinaryBuffer = function () {
        var bufferArray = [];
        this.forEach(function (slice) {
            bufferArray = bufferArray.concat(slice.asBinaryBuffer());
        });
        return bufferArray;
    };
    return TimeboxEvoMessageArray;
}(Array));

function unhexlify(str) {
    var result = '';
    if (str.length % 2 !== 0) {
        throw new Error('The string length is not a multiple of 2');
    }
    for (var i = 0, l = str.length; i < l; i += 2) {
        var toHex = parseInt(str.substr(i, 2), 16);
        if (isNaN(toHex)) {
            throw new Error('str contains non hex character');
        }
        result += String.fromCharCode(toHex);
    }
    return result;
}
function int2hexlittle(value) {
    if (value > 65535 || value < 0) {
        throw new TypeError('int2hexlittle only supports value between 0 and 65535');
    }
    var byte1 = (value & 0xFF).toString(16).padStart(2, "0");
    var byte2 = ((value >> 8) & 0xFF).toString(16).padStart(2, "0");
    return "" + byte1 + byte2;
}
function number2HexString(int) {
    if (int > 255 || int < 0) {
        throw new Error('number2HexString works only with number between 0 and 255');
    }
    return Math.round(int).toString(16).padStart(2, "0");
}
function boolean2HexString(bool) {
    return bool ? "01" : "00";
}
function color2HexString(color) {
    return new TinyColor(color).toHex();
}
function brightness2HexString(brightness) {
    if (brightness > 100) {
        brightness = 100;
    }
    else if (brightness < 0) {
        brightness = 0;
    }
    return number2HexString(Math.trunc(brightness));
}

var TimeboxEvoMessage = (function () {
    function TimeboxEvoMessage(msg) {
        if (msg === void 0) { msg = ''; }
        this._START = "01";
        this._END = "02";
        this.append(msg);
    }
    TimeboxEvoMessage.prototype._calcCRC = function () {
        if (!this._message)
            return undefined;
        var msg = this.lengthHS + this._message;
        var sum = 0;
        for (var i = 0, l = msg.length; i < l; i += 2) {
            sum += parseInt(msg.substr(i, 2), 16);
        }
        return sum % 65536;
    };
    Object.defineProperty(TimeboxEvoMessage.prototype, "crc", {
        get: function () {
            if (!this._message)
                return undefined;
            return this._calcCRC();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeboxEvoMessage.prototype, "crcHS", {
        get: function () {
            if (!this._message)
                return undefined;
            return int2hexlittle(this.crc);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeboxEvoMessage.prototype, "length", {
        get: function () {
            if (!this._message)
                return undefined;
            return (this._message.length + 4) / 2;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeboxEvoMessage.prototype, "lengthHS", {
        get: function () {
            if (!this._message)
                return undefined;
            return int2hexlittle(this.length);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeboxEvoMessage.prototype, "payload", {
        get: function () {
            return this._message;
        },
        set: function (payload) {
            this._message = payload;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeboxEvoMessage.prototype, "message", {
        get: function () {
            if (!this._message)
                return undefined;
            return this._START + this.lengthHS + this._message + this.crcHS + this._END;
        },
        enumerable: true,
        configurable: true
    });
    TimeboxEvoMessage.prototype.append = function (msg) {
        if (msg) {
            this._message = this._message ? this._message + msg.toLowerCase() : msg.toLowerCase();
        }
        return this;
    };
    TimeboxEvoMessage.prototype.prepend = function (msg) {
        if (msg) {
            this._message = this._message ? msg.toLowerCase() + this._message : msg.toLowerCase();
        }
        return this;
    };
    TimeboxEvoMessage.prototype.toString = function () {
        return this.message;
    };
    TimeboxEvoMessage.prototype.asBinaryBuffer = function () {
        var bufferArray = [];
        this.message.match(/.{1,1332}/g).forEach(function (part) {
            bufferArray.push(Buffer.from(unhexlify(part), 'binary'));
        });
        return bufferArray;
    };
    return TimeboxEvoMessage;
}());

var TimeboxEvoRequest = (function () {
    function TimeboxEvoRequest() {
        this._messages = TimeboxEvoMessageArray.create();
    }
    TimeboxEvoRequest.prototype.push = function (msg) {
        return this._messages.push(new TimeboxEvoMessage(msg));
    };
    TimeboxEvoRequest.prototype.clear = function () {
        this._messages = TimeboxEvoMessageArray.create();
    };
    Object.defineProperty(TimeboxEvoRequest.prototype, "messages", {
        get: function () {
            return this._messages;
        },
        enumerable: true,
        configurable: true
    });
    return TimeboxEvoRequest;
}());

var TimeDisplayType;
(function (TimeDisplayType) {
    TimeDisplayType[TimeDisplayType["FullScreen"] = 0] = "FullScreen";
    TimeDisplayType[TimeDisplayType["Rainbow"] = 1] = "Rainbow";
    TimeDisplayType[TimeDisplayType["WithBox"] = 2] = "WithBox";
    TimeDisplayType[TimeDisplayType["AnalogSquare"] = 3] = "AnalogSquare";
    TimeDisplayType[TimeDisplayType["FullScreenNegative"] = 4] = "FullScreenNegative";
    TimeDisplayType[TimeDisplayType["AnalogRound"] = 5] = "AnalogRound";
})(TimeDisplayType || (TimeDisplayType = {}));
var LightningType;
(function (LightningType) {
    LightningType[LightningType["PlainColor"] = 0] = "PlainColor";
    LightningType[LightningType["Love"] = 1] = "Love";
    LightningType[LightningType["Plants"] = 2] = "Plants";
    LightningType[LightningType["NoMosquitto"] = 3] = "NoMosquitto";
    LightningType[LightningType["Sleeping"] = 4] = "Sleeping";
})(LightningType || (LightningType = {}));
var WeatherType;
(function (WeatherType) {
    WeatherType[WeatherType["Clear"] = 1] = "Clear";
    WeatherType[WeatherType["CloudySky"] = 3] = "CloudySky";
    WeatherType[WeatherType["Thunderstorm"] = 5] = "Thunderstorm";
    WeatherType[WeatherType["Rain"] = 6] = "Rain";
    WeatherType[WeatherType["Snow"] = 8] = "Snow";
    WeatherType[WeatherType["Fog"] = 9] = "Fog";
})(WeatherType || (WeatherType = {}));
var VJEffectType;
(function (VJEffectType) {
    VJEffectType[VJEffectType["Sparkles"] = 0] = "Sparkles";
    VJEffectType[VJEffectType["Lava"] = 1] = "Lava";
    VJEffectType[VJEffectType["VerticalRainbowLines"] = 2] = "VerticalRainbowLines";
    VJEffectType[VJEffectType["Drops"] = 3] = "Drops";
    VJEffectType[VJEffectType["RainbowSwirl"] = 4] = "RainbowSwirl";
    VJEffectType[VJEffectType["CMYFade"] = 5] = "CMYFade";
    VJEffectType[VJEffectType["RainbowLava"] = 6] = "RainbowLava";
    VJEffectType[VJEffectType["PastelPatterns"] = 7] = "PastelPatterns";
    VJEffectType[VJEffectType["CMYWave"] = 8] = "CMYWave";
    VJEffectType[VJEffectType["Fire"] = 9] = "Fire";
    VJEffectType[VJEffectType["Countdown"] = 10] = "Countdown";
    VJEffectType[VJEffectType["PinkBlueFade"] = 11] = "PinkBlueFade";
    VJEffectType[VJEffectType["RainbowPolygons"] = 12] = "RainbowPolygons";
    VJEffectType[VJEffectType["PinkBlueWave"] = 13] = "PinkBlueWave";
    VJEffectType[VJEffectType["RainbowCross"] = 14] = "RainbowCross";
    VJEffectType[VJEffectType["RainbowShapes"] = 15] = "RainbowShapes";
})(VJEffectType || (VJEffectType = {}));
var TIMEBOX_CONST = {
    TimeType: TimeDisplayType,
    LightningType: LightningType,
    WeatherType: WeatherType,
    VJEffectType: VJEffectType
};

var TimeChannel = (function (_super) {
    __extends(TimeChannel, _super);
    function TimeChannel(opts) {
        var _this = _super.call(this) || this;
        _this._opts = {
            type: TimeDisplayType.FullScreen,
            showTime: true,
            showWeather: false,
            showTemp: false,
            showCalendar: false
        };
        _this._PACKAGE_PREFIX = "450001";
        _this.color =
            opts && opts.color ? new TinyColor(opts.color) : new TinyColor("FFFFFF");
        _this._opts = __assign(__assign({}, _this._opts), opts);
        _this._updateMessage();
        return _this;
    }
    TimeChannel.prototype._updateMessage = function () {
        this.clear();
        this.push(this._PACKAGE_PREFIX +
            number2HexString(this._opts.type) +
            boolean2HexString(this._opts.showTime) +
            boolean2HexString(this._opts.showWeather) +
            boolean2HexString(this._opts.showTemp) +
            boolean2HexString(this._opts.showCalendar) +
            color2HexString(this._color));
    };
    Object.defineProperty(TimeChannel.prototype, "type", {
        get: function () {
            return this._opts.type;
        },
        set: function (type) {
            this._opts.type = type;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeChannel.prototype, "color", {
        get: function () {
            return this._color;
        },
        set: function (color) {
            var localcolor = new TinyColor(color);
            if (!localcolor.isValid)
                throw new Error("Provided color " + localcolor + " is not valid");
            this._color = localcolor.toHex();
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeChannel.prototype, "showTime", {
        get: function () {
            return this._opts.showTime;
        },
        set: function (bool) {
            this._opts.showTime = bool;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeChannel.prototype, "showWeather", {
        get: function () {
            return this._opts.showWeather;
        },
        set: function (bool) {
            this._opts.showWeather = bool;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeChannel.prototype, "showTemp", {
        get: function () {
            return this._opts.showTemp;
        },
        set: function (bool) {
            this._opts.showTemp = bool;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TimeChannel.prototype, "showCalendar", {
        get: function () {
            return this._opts.showCalendar;
        },
        set: function (bool) {
            this._opts.showCalendar = bool;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    return TimeChannel;
}(TimeboxEvoRequest));

var LightningChannel = (function (_super) {
    __extends(LightningChannel, _super);
    function LightningChannel(opts) {
        var _this = _super.call(this) || this;
        _this._opts = {
            type: LightningType.PlainColor,
            brightness: 100,
            power: true
        };
        _this._PACKAGE_PREFIX = "4501";
        _this._PACKAGE_SUFFIX = "000000";
        _this.color =
            opts && opts.color ? new TinyColor(opts.color) : new TinyColor("FFFFFF");
        _this._opts = __assign(__assign({}, _this._opts), opts);
        _this._updateMessage();
        return _this;
    }
    LightningChannel.prototype._updateMessage = function () {
        this.clear();
        this.push(this._PACKAGE_PREFIX +
            color2HexString(this._color) +
            brightness2HexString(this._opts.brightness) +
            number2HexString(this._opts.type) +
            boolean2HexString(this._opts.power) +
            this._PACKAGE_SUFFIX);
    };
    Object.defineProperty(LightningChannel.prototype, "type", {
        get: function () {
            return this._opts.type;
        },
        set: function (type) {
            this._opts.type = type;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(LightningChannel.prototype, "color", {
        get: function () {
            return this._color;
        },
        set: function (color) {
            var localcolor = new TinyColor(color);
            if (!localcolor.isValid)
                throw new Error("Provided color " + localcolor + " is not valid");
            this._color = localcolor.toHex();
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(LightningChannel.prototype, "power", {
        get: function () {
            return this._opts.power;
        },
        set: function (bool) {
            this._opts.power = bool;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(LightningChannel.prototype, "brightness", {
        get: function () {
            return this._opts.brightness;
        },
        set: function (brightness) {
            this._opts.brightness = brightness;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    return LightningChannel;
}(TimeboxEvoRequest));

var CloudChannel = (function (_super) {
    __extends(CloudChannel, _super);
    function CloudChannel() {
        var _this = _super.call(this) || this;
        _this.push("4502");
        return _this;
    }
    return CloudChannel;
}(TimeboxEvoRequest));

var VJEffectChannel = (function (_super) {
    __extends(VJEffectChannel, _super);
    function VJEffectChannel(opts) {
        var _this = _super.call(this) || this;
        _this._opts = {
            type: 0
        };
        _this._PACKAGE_PREFIX = "4503";
        _this._opts = __assign(__assign({}, _this._opts), opts);
        _this._updateMessage();
        return _this;
    }
    VJEffectChannel.prototype._updateMessage = function () {
        this.clear();
        this.push(this._PACKAGE_PREFIX + number2HexString(this._opts.type));
    };
    Object.defineProperty(VJEffectChannel.prototype, "type", {
        get: function () {
            return this._opts.type;
        },
        set: function (type) {
            this._opts.type = type;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    return VJEffectChannel;
}(TimeboxEvoRequest));

var ScoreBoardChannel = (function (_super) {
    __extends(ScoreBoardChannel, _super);
    function ScoreBoardChannel(opts) {
        var _this = _super.call(this) || this;
        _this._opts = {
            red: 0,
            blue: 0
        };
        _this._PACKAGE_PREFIX = "450600";
        _this.red = opts && opts.red ? opts.red : 0;
        _this.blue = opts && opts.blue ? opts.blue : 0;
        return _this;
    }
    ScoreBoardChannel.prototype._updateMessage = function () {
        this.clear();
        this.push(this._PACKAGE_PREFIX +
            int2hexlittle(this._opts.red) +
            int2hexlittle(this._opts.blue));
    };
    Object.defineProperty(ScoreBoardChannel.prototype, "red", {
        get: function () {
            return this._opts.red;
        },
        set: function (int) {
            this._opts.red = Math.min(999, Math.max(0, int));
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ScoreBoardChannel.prototype, "blue", {
        get: function () {
            return this._opts.blue;
        },
        set: function (int) {
            this._opts.blue = Math.min(999, Math.max(0, int));
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    return ScoreBoardChannel;
}(TimeboxEvoRequest));

var CustomChannel = (function (_super) {
    __extends(CustomChannel, _super);
    function CustomChannel() {
        var _this = _super.call(this) || this;
        _this._PACKAGE_HEADER = "4505";
        _this.push(_this._PACKAGE_HEADER);
        return _this;
    }
    return CustomChannel;
}(TimeboxEvoRequest));

var BrightnessCommand = (function (_super) {
    __extends(BrightnessCommand, _super);
    function BrightnessCommand(opts) {
        var _this = _super.call(this) || this;
        _this._opts = {
            brightness: 100,
            in_min: 0,
            in_max: 100
        };
        _this._PACKAGE_PREFIX = "74";
        _this._opts = __assign(__assign({}, _this._opts), opts);
        _this._updateMessage();
        return _this;
    }
    Object.defineProperty(BrightnessCommand.prototype, "brightness", {
        get: function () {
            return this._opts.brightness;
        },
        set: function (brightness) {
            this._opts.brightness = brightness;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BrightnessCommand.prototype, "in_min", {
        get: function () {
            return this._opts.in_min;
        },
        set: function (in_min) {
            this._opts.in_min = in_min;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BrightnessCommand.prototype, "in_max", {
        get: function () {
            return this._opts.in_max;
        },
        set: function (in_max) {
            this._opts.in_max = in_max;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(BrightnessCommand.prototype, "opts", {
        get: function () {
            return this._opts;
        },
        set: function (opts) {
            this._opts = __assign(__assign({}, this._opts), opts);
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    BrightnessCommand.prototype._updateMessage = function () {
        function map(x, in_min, in_max, out_min, out_max) {
            if (x < in_min || x > in_max) {
                throw new Error("map() in_min is < value or in_max > value");
            }
            return ((x - in_min) * (out_max - out_min)) / (in_max - in_min) + out_min;
        }
        if ((this._opts.brightness > 100 || this._opts.brightness < 0) &&
            (this._opts.in_min === undefined || this._opts.in_max === undefined)) {
            throw new Error("Brightness should be between 0 and 100 or in_min and in_max should be defined");
        }
        var briInRange = this._opts.brightness;
        if (this._opts.in_min !== undefined && this._opts.in_max !== undefined) {
            briInRange = Math.ceil(map(this._opts.brightness, this._opts.in_min, this._opts.in_max, 0, 100));
        }
        this.clear();
        this.push(this._PACKAGE_PREFIX + number2HexString(briInRange));
    };
    return BrightnessCommand;
}(TimeboxEvoRequest));

var TempWeatherCommand = (function (_super) {
    __extends(TempWeatherCommand, _super);
    function TempWeatherCommand(opts) {
        var _this = _super.call(this) || this;
        _this._PACKAGE_PREFIX = "5F";
        _this._opts = {
            temperature: 0,
            weather: TIMEBOX_CONST.WeatherType.Clear
        };
        _this._opts = __assign(__assign({}, _this._opts), opts);
        _this.temperature = _this._opts.temperature;
        _this.weather = _this._opts.weather;
        return _this;
    }
    Object.defineProperty(TempWeatherCommand.prototype, "temperature", {
        get: function () {
            return this._opts.temperature;
        },
        set: function (temp) {
            if (temp > 128 || temp < -127) {
                throw new Error("temp should be >= -127 and <= 128");
            }
            this._opts.temperature = temp;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TempWeatherCommand.prototype, "weather", {
        get: function () {
            return this._opts.weather;
        },
        set: function (weather) {
            this._opts.weather = weather;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    TempWeatherCommand.prototype._updateMessage = function () {
        this.clear();
        var encodedTemp = "";
        if (this._opts.temperature >= 0) {
            encodedTemp = number2HexString(this._opts.temperature);
        }
        else {
            var value = 256 + this._opts.temperature;
            encodedTemp = number2HexString(value);
        }
        this.push(this._PACKAGE_PREFIX + encodedTemp + number2HexString(this._opts.weather));
    };
    return TempWeatherCommand;
}(TimeboxEvoRequest));

var DateTimeCommand = (function (_super) {
    __extends(DateTimeCommand, _super);
    function DateTimeCommand(opts) {
        var _this = _super.call(this) || this;
        _this._PACKAGE_PREFIX = "18";
        _this._opts = {
            date: new Date(),
        };
        _this._opts = __assign(__assign({}, _this._opts), opts);
        _this._updateMessage();
        return _this;
    }
    Object.defineProperty(DateTimeCommand.prototype, "date", {
        get: function () {
            return this._opts.date;
        },
        set: function (date) {
            this._opts.date = date;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    DateTimeCommand.prototype._updateMessage = function () {
        this.clear();
        var timeString = number2HexString(Number(this._opts.date.getFullYear().toString().padStart(4, "0").slice(2)))
            + number2HexString(Number(this._opts.date.getFullYear().toString().padStart(4, "0").slice(0, 2)))
            + number2HexString(this._opts.date.getMonth() + 1)
            + number2HexString(this._opts.date.getDate())
            + number2HexString(this._opts.date.getHours())
            + number2HexString(this._opts.date.getMinutes())
            + number2HexString(this._opts.date.getSeconds())
            + "00";
        this.push(this._PACKAGE_PREFIX + timeString);
    };
    return DateTimeCommand;
}(TimeboxEvoRequest));

var DisplayText = (function (_super) {
    __extends(DisplayText, _super);
    function DisplayText(opts) {
        var _this = _super.call(this) || this;
        _this._animFrame = 0;
        _this._opts = {
            text: "node-divoom-timebox-evo",
            paletteFn: _this.PALETTE_TEXT_ON_BACKGROUND,
            animFn: _this.ANIM_STATIC_BACKGROUND
        };
        _this._opts = __assign(__assign({}, _this._opts), opts);
        _this._updateMessage();
        return _this;
    }
    DisplayText.prototype.PALETTE_TEXT_ON_BACKGROUND = function (text, background) {
        if (text === void 0) { text = "FFFFFF"; }
        if (background === void 0) { background = "000000"; }
        var back = (new TinyColor(background)).toHex();
        var front = (new TinyColor(text)).toHex();
        var palette = Array.from({ length: 256 }).fill(back, 0, 127);
        palette.fill(front, 127);
        return palette;
    };
    DisplayText.prototype.PALETTE_BLACK_ON_CMY_RAINBOW = function () {
        var palette = [];
        var r = 255, g = 0, b = 255;
        for (var i = 0; i < 254; i += 2) {
            palette.push(number2HexString(r) + number2HexString(g) + number2HexString(b));
            if (i < 85) {
                b = Math.max(0, b - 6);
                g = Math.min(255, g + 6);
            }
            else if (i < 170) {
                b = Math.min(255, b + 6);
                r = Math.max(0, r - 6);
            }
            else {
                r = Math.min(255, r + 6);
                g = Math.max(0, g - 6);
            }
        }
        for (var i = palette.length; i < 256; i++) {
            palette.push("000000");
        }
        return palette;
    };
    DisplayText.prototype.PALETTE_BLACK_ON_RAINBOW = function () {
        var palette = [];
        var size = 127;
        function sin_to_hex(i, phase) {
            var sin = Math.sin(Math.PI / size * 2 * i + phase);
            var int = Math.floor(sin * 127) + 128;
            return number2HexString(int);
        }
        for (var i = 0; i < size; i++) {
            var red = sin_to_hex(i, 0 * Math.PI * 2 / 3);
            var blue = sin_to_hex(i, 1 * Math.PI * 2 / 3);
            var green = sin_to_hex(i, 2 * Math.PI * 2 / 3);
            palette.push(red + green + blue);
        }
        for (var i = palette.length; i < 256; i++) {
            palette.push("000000");
        }
        return palette;
    };
    DisplayText.prototype.ANIM_STATIC_BACKGROUND = function (i) {
        return Array.from({ length: 256 }).fill(0);
    };
    DisplayText.prototype.ANIM_UNI_GRADIANT_BACKGROUND = function (i) {
        return Array.from({ length: 256 }).fill(i % 127);
    };
    DisplayText.prototype.ANIM_HORIZONTAL_GRADIANT_BACKGROUND = function (frame) {
        var pixelArray = [];
        for (var y = 0; y < 16; y++) {
            for (var x = 0; x < 16; x++) {
                pixelArray.push((x + frame) % 127);
            }
        }
        return pixelArray;
    };
    DisplayText.prototype.ANIM_VERTICAL_GRADIANT_BACKGROUND = function (frame) {
        var pixelArray = [];
        for (var y = 0; y < 16; y++) {
            for (var x = 0; x < 16; x++) {
                pixelArray.push((y + frame) % 127);
            }
        }
        return pixelArray;
    };
    DisplayText.prototype._encodeText = function (text) {
        var length = number2HexString(text.length);
        var encodedText = "8601" + length;
        text.split("").forEach(function (char) {
            encodedText += int2hexlittle(char.charCodeAt(0));
        });
        return encodedText;
    };
    DisplayText.prototype._updateMessage = function () {
        this.clear();
        var PACKAGE_INIT_MESSAGE = "6e01";
        if (typeof this._opts.animFn !== 'function' || typeof this._opts.paletteFn !== 'function') {
            throw new Error('paletteFn and animFn need to be functions');
        }
        this._animFrame = 0;
        this.push(PACKAGE_INIT_MESSAGE);
        this.push(this._encodeText(this._opts.text));
        var PALETTE_HEADER = "6c00000704aa070446000000";
        var pixels = '';
        var palette = this.colorPalette;
        this._opts.animFn(this._animFrame).forEach(function (pixel) {
            pixels += number2HexString(pixel);
        });
        this._animFrame++;
        this.push(PALETTE_HEADER
            + palette.join("")
            + pixels);
        this.push(this.getNextAnimationFrame().payload);
    };
    DisplayText.prototype.getNextAnimationFrame = function () {
        var pixelArray = this._opts.animFn(this._animFrame);
        if (pixelArray.length !== 256)
            throw new Error('The animFn should always generate a 256 pixel array');
        var pixelString = '';
        pixelArray.forEach(function (pixel) {
            pixelString += number2HexString(pixel);
        });
        var animString = "6c"
            + int2hexlittle(this._animFrame)
            + "0701aa070143000100"
            + pixelString;
        this._animFrame = ++this._animFrame % 65536;
        return new TimeboxEvoMessage(animString);
    };
    Object.defineProperty(DisplayText.prototype, "paletteFn", {
        get: function () {
            return this._opts.paletteFn;
        },
        set: function (paletteFn) {
            if (typeof paletteFn !== 'function') {
                throw new Error('paletteFn is not a function');
            }
            this._opts.paletteFn = paletteFn;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayText.prototype, "colorPalette", {
        get: function () {
            var palette = this.paletteFn();
            if (palette.length !== 256) {
                throw new Error('The paletteFn should always generate 256 colors');
            }
            var result = [];
            palette.forEach(function (color) {
                var lColor = new TinyColor(color);
                if (!lColor.isValid) {
                    throw new Error('One of your color is not valid');
                }
                result.push(lColor.toHex());
            });
            return result;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayText.prototype, "animFn", {
        get: function () {
            return this._opts.animFn;
        },
        set: function (animFn) {
            if (typeof animFn !== 'function') {
                throw new Error('paletteFn is not a function');
            }
            this._opts.animFn = animFn;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayText.prototype, "pixels", {
        get: function () {
            return this.animFn(this._animFrame);
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayText.prototype, "frame", {
        get: function () {
            return this._animFrame;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DisplayText.prototype, "text", {
        get: function () {
            return this._opts.text;
        },
        set: function (text) {
            this._opts.text = text;
            this._updateMessage();
        },
        enumerable: true,
        configurable: true
    });
    return DisplayText;
}(TimeboxEvoRequest));

var JimpArray = (function (_super) {
    __extends(JimpArray, _super);
    function JimpArray(items) {
        return _super.apply(this, items) || this;
    }
    JimpArray.create = function () {
        return Object.create(JimpArray.prototype);
    };
    JimpArray.prototype._animAsDivoomMessages = function () {
        var _PACKAGE_PREFIX = '49';
        var dms = TimeboxEvoMessageArray.create();
        var fullString = '';
        var totalSize = 0;
        this.forEach(function (image) {
            var dm = image.asDivoomMessage();
            fullString += dm.payload;
            totalSize += dm.payload.length / 2;
        });
        var messageCounter = 0;
        var totalSizeHex = int2hexlittle(totalSize);
        fullString.match(/.{1,400}/g).forEach(function (message) {
            dms.push(new TimeboxEvoMessage(_PACKAGE_PREFIX
                + totalSizeHex
                + number2HexString(messageCounter)
                + message));
            messageCounter++;
        });
        return dms;
    };
    JimpArray.prototype._staticAsDivoomMessages = function () {
        var PACKAGE_PREFIX = '44000A0A04';
        var dms = TimeboxEvoMessageArray.create();
        dms.push(this[0].asDivoomMessage().prepend(PACKAGE_PREFIX));
        return dms;
    };
    JimpArray.prototype.asDivoomMessages = function () {
        if (this[0]) {
            if (this[0] instanceof DivoomJimpAnim) {
                return this._animAsDivoomMessages();
            }
            else if (this[0] instanceof DivoomJimpStatic) {
                return this._staticAsDivoomMessages();
            }
        }
        else {
            return TimeboxEvoMessageArray.create();
        }
    };
    JimpArray.prototype.asBinaryBuffer = function () {
        var bufferArray = [];
        this.asDivoomMessages().forEach(function (message) {
            bufferArray = bufferArray.concat(message.asBinaryBuffer());
        });
        return bufferArray;
    };
    return JimpArray;
}(Array));
var DivoomJimp = (function (_super) {
    __extends(DivoomJimp, _super);
    function DivoomJimp() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    DivoomJimp.prototype.getColorsAndPixels = function () {
        var colorsArray = [];
        var pixelArray = [];
        var colorCounter = 0;
        this.scan(0, 0, this.bitmap.width, this.bitmap.height, function (x, y, idx) {
            var red = this.bitmap.data[idx + 0];
            var green = this.bitmap.data[idx + 1];
            var blue = this.bitmap.data[idx + 2];
            var color = red.toString(16).padStart(2, "0")
                + green.toString(16).padStart(2, "0")
                + blue.toString(16).padStart(2, "0");
            if (!colorsArray.includes(color)) {
                colorsArray.push(color);
                pixelArray[x + 16 * y] = colorCounter;
                colorCounter++;
            }
            else {
                pixelArray[x + 16 * y] = colorsArray.indexOf(color);
            }
        });
        return { colors: colorsArray, pixels: pixelArray };
    };
    DivoomJimp.prototype.getPixelString = function (pixelArray, nbColors) {
        var nbBitsForAPixel = Math.log(nbColors) / Math.log(2);
        var bits = Number.isInteger(nbBitsForAPixel)
            ? nbBitsForAPixel
            : (Math.trunc(nbBitsForAPixel) + 1);
        if (bits === 0)
            bits = 1;
        var pixelString = '';
        pixelArray.forEach(function (pixel) {
            pixelString += pixel.toString(2).padStart(8, '0').split("").reverse().join("").substring(0, bits);
        });
        var pixBinArray = pixelString.match(/.{1,8}/g);
        var pixelStringFinal = '';
        pixBinArray.forEach(function (pixel) {
            pixelStringFinal += parseInt(pixel.split("").reverse().join(""), 2).toString(16).padStart(2, '0');
        });
        return pixelStringFinal;
    };
    DivoomJimp.prototype.asBinaryBuffer = function () {
        return [];
    };
    return DivoomJimp;
}(Jimp));
var DivoomJimpAnim = (function (_super) {
    __extends(DivoomJimpAnim, _super);
    function DivoomJimpAnim() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._frame = 0;
        _this._delay = 0;
        return _this;
    }
    Object.defineProperty(DivoomJimpAnim.prototype, "frame", {
        get: function () {
            return this._frame;
        },
        set: function (frame) {
            this._frame = frame;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DivoomJimpAnim.prototype, "delay", {
        get: function () {
            return this._delay;
        },
        set: function (delay) {
            this._delay = delay;
        },
        enumerable: true,
        configurable: true
    });
    DivoomJimpAnim.prototype.asDivoomMessage = function () {
        var colorsAndPixels = this.getColorsAndPixels();
        var nbColorsHex = number2HexString(colorsAndPixels.colors.length % 256);
        var colorString = colorsAndPixels.colors.join("");
        var pixelString = this.getPixelString(colorsAndPixels.pixels, colorsAndPixels.colors.length);
        var delayHex = int2hexlittle(this.delay);
        var stringWithoutHeader = delayHex
            + ( "00" )
            + nbColorsHex
            + colorString
            + pixelString;
        var sizeHex = int2hexlittle((stringWithoutHeader.length + 6) / 2);
        var fullString = 'aa' +
            sizeHex +
            stringWithoutHeader;
        return new TimeboxEvoMessage(fullString);
    };
    return DivoomJimpAnim;
}(DivoomJimp));
var DivoomJimpStatic = (function (_super) {
    __extends(DivoomJimpStatic, _super);
    function DivoomJimpStatic() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    DivoomJimpStatic.prototype.asDivoomMessage = function () {
        var colorsAndPixels = this.getColorsAndPixels();
        var nbColorsHex = number2HexString(colorsAndPixels.colors.length % 256);
        var colorString = colorsAndPixels.colors.join("");
        var pixelString = this.getPixelString(colorsAndPixels.pixels, colorsAndPixels.colors.length);
        var stringWithoutHeader = nbColorsHex
            + colorString
            + pixelString;
        var sizeHex = int2hexlittle((('AA0000000000' + stringWithoutHeader).length) / 2);
        var fullString = 'aa'
            + sizeHex
            + '000000'
            + stringWithoutHeader;
        return new TimeboxEvoMessage(fullString);
    };
    return DivoomJimpStatic;
}(DivoomJimp));

var DisplayAnimation = (function (_super) {
    __extends(DisplayAnimation, _super);
    function DisplayAnimation() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    DisplayAnimation.prototype.read = function (input) {
        return __awaiter(this, void 0, void 0, function () {
            var buffer, ft, _a;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        if (Buffer.isBuffer(input)) {
                            buffer = input;
                        }
                        else {
                            buffer = fs.readFileSync(input);
                        }
                        ft = fileType(buffer);
                        ft = fileType(buffer);
                        if (!ft) return [3, 7];
                        _a = ft.mime;
                        switch (_a) {
                            case "image/gif": return [3, 1];
                            case "image/jpeg": return [3, 3];
                            case "image/png": return [3, 3];
                            case "image/bmp": return [3, 3];
                        }
                        return [3, 5];
                    case 1: return [4, this._displayAnimationFromGIF(buffer)];
                    case 2: return [2, _b.sent()];
                    case 3: return [4, this._displayImage(buffer)];
                    case 4: return [2, _b.sent()];
                    case 5: throw new Error("file type not supported");
                    case 6: return [3, 8];
                    case 7: throw new Error("file type unkown");
                    case 8: return [2];
                }
            });
        });
    };
    DisplayAnimation.prototype._displayImage = function (input) {
        return __awaiter(this, void 0, void 0, function () {
            var ja, image, resized;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        ja = JimpArray.create();
                        return [4, Jimp.read(input)];
                    case 1:
                        image = _a.sent();
                        resized = new DivoomJimpStatic(image.resize(16, 16, Jimp.RESIZE_NEAREST_NEIGHBOR));
                        ja.push(resized);
                        return [2, ja];
                }
            });
        });
    };
    DisplayAnimation.prototype._displayAnimationFromGIF = function (input) {
        return __awaiter(this, void 0, void 0, function () {
            var gifCodec, inputGif, ja;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        gifCodec = new gifWrap.GifCodec();
                        return [4, gifCodec.decodeGif(input)];
                    case 1:
                        inputGif = _a.sent();
                        ja = JimpArray.create();
                        inputGif.frames.forEach(function (frame, index) {
                            var image = gifWrap.GifUtil.copyAsJimp(DivoomJimpAnim, frame).resize(16, 16);
                            image.delay = frame.delayCentisecs * 10;
                            image.frame = index;
                            ja.push(image);
                        });
                        return [2, ja];
                }
            });
        });
    };
    return DisplayAnimation;
}(TimeboxEvoRequest));

var TimeboxEvo = (function () {
    function TimeboxEvo() {
    }
    TimeboxEvo.prototype.createRequest = function (type, opts) {
        switch (type.toLowerCase()) {
            case "cloud":
                return new CloudChannel();
            case "custom":
                return new CustomChannel();
            case "lightning":
                return new LightningChannel(opts);
            case "scoreboard":
                return new ScoreBoardChannel(opts);
            case "time":
                return new TimeChannel(opts);
            case "vj-effect":
            case "vjeffect":
                return new VJEffectChannel(opts);
            case "brightness":
                return new BrightnessCommand(opts);
            case "temp_weather":
                return new TempWeatherCommand(opts);
            case "text":
                return new DisplayText(opts);
            case "picture":
            case "animation":
                return new DisplayAnimation();
            case "datetime":
                return new DateTimeCommand(opts);
            case "raw":
                return new TimeboxEvoRequest();
            default:
                throw new Error("Unkown type");
        }
    };
    return TimeboxEvo;
}());

export { BrightnessCommand, CloudChannel, CustomChannel, DateTimeCommand, DisplayAnimation, DisplayText, LightningChannel, ScoreBoardChannel, TIMEBOX_CONST, TempWeatherCommand, TimeChannel, TimeboxEvo, TimeboxEvoRequest, VJEffectChannel };
